#! /usr/bin/python
# Copyright 2016 Cumulus Networks , Inc.

#############################################################################
# Platform and model specific poe subclass, inherits from the base poe class,
# and provides the followings:
# - loads the chip-specific poe implementation
#

try:
    import exceptions
    import os
    import termios
    import time
    import select
    import errno
    import fcntl
    import struct
    import serial
    import poe_base
    import poe_pd69100
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

def poe():
    return poe_dni3048()

class poe_dni3048(poe_pd69100.PoePD69100):
    ''' according to dni3048 spec. POE port settings are:
          port             - /dev/ttyS1
          baud rate        - 19200 bps
          stop bits        - 1
          parity           - none
          data bits        - 8
          flow control     - none
          interpacket gap  - > 20ms
          timeout          - 50ms
          byte gap         - < 2ms
    '''
    def __init__(self):
        poe_pd69100.PoePD69100.__init__(self)
        self.log_debug('poe_dni3048t::init()')
        self._fd = None

    def poe_enable(self):
        self._enable_poe_in_cpld()

    def handle_warm_cold_boot(self):

        self.set_individual_mask(poe_pd69100.PD69100_IND_MASK_4PAIR_ENABLE, 1)
        port_map = {
             0:( 7,  6),  1:( 5,  4),  2:( 3, 2),   3:( 1,  0),
             4:(15, 14),  5:(13, 12),  6:(11, 10),  7:( 9,  8),
             8:(23, 22),  9:(21, 20), 10:(19, 18), 11:(17, 16),
            12:(31, 30), 13:(29, 28), 14:(27, 26), 15:(25, 24),
            16:(39, 38), 17:(37, 36), 18:(35, 34), 19:(33, 32),
            20:(47, 46), 21:(45, 44), 22:(43, 42), 23:(41, 40),
            24:(55, 54), 25:(53, 52), 26:(51, 50), 27:(49, 48),
            28:(63, 62), 29:(61, 60), 30:(59, 58), 31:(57, 56),
            32:(71, 70), 33:(69, 68), 34:(67, 66), 35:(65, 64),
            36:(79, 78), 37:(77, 76), 38:(75, 74), 39:(73, 72),
            40:(87, 86), 41:(85, 84), 42:(83, 82), 43:(81, 80),
            44:(95, 94), 45:(93, 92), 46:(91, 90), 47:(89, 88)
        }
        for port_id, port_phys in port_map.items():
            self.set_port_temp_matrix(port_id, port_phys[0], port_phys[1])
        self.set_active_matrix()

        self.set_port_disable_poe(range(48))

        ''' turn on L2 management
        '''
        self.set_individual_mask(poe_pd69100.PD69100_IND_MASK_LAYER2_LLDP, 1)
        self.set_pm_settings(poe_pd69100.PD69100_PM_MODE_PM1_LLDP,
                             poe_pd69100.PD69100_PM_MODE_PM2_PPL,
                             poe_pd69100.PD69100_PM_MODE_PM3_NONE)

        time.sleep(0.1)
        ''' enable 4 pair on all ports
        '''
        self.set_individual_mask(poe_pd69100.PD69100_IND_MASK_4PAIR_ENABLE, 1)
        self.set_enable_4pair_mode(range(48))
        time.sleep(0.1)
        self.set_disable_4pair_power(range(48))

        self._persist_settings()

        ''' this will apply the saved configuration
        '''
        poe_pd69100.PoePD69100.handle_warm_cold_boot(self)

    def handle_poed_start(self):
        poe_pd69100.PoePD69100.handle_poed_start(self)
        '''
            enable POE in the CPLD
        '''
        self.enable_poe_in_cpld()

    @property
    def fd(self):
        if not self._fd:
            self._fd = serial.Serial('/dev/ttyS1', baudrate=19200, bytesize=8, parity='N', stopbits=1, timeout=1)
        return self._fd

    def enable_poe_in_cpld(self):
        open('/sys/bus/i2c/devices/0-0028/poe_plus_power', 'w').write('1')

    def disable_poe_in_cpld(self):
        open('/sys/bus/i2c/devices/0-0028/poe_plus_power', 'w').write('0')

    def poe_write(self, data):
        self.log_debug('poe_dni3048t::write()')
        return self.fd.write(struct.pack('B' * len(data), *data))

    def poe_read(self, size=1):
        self.log_debug('poe_dni3048t::read()')
        read_data = self.fd.read(size)
        return struct.unpack('B' * len(read_data), read_data)

    def get_poe_params(self):
        ports = {}
        for port_id in range(48):
            ports[port_id] = poe_base.PoePortParams(port_id, 'swp%u' % (port_id + 1),
                                                    True, poe_base.MAX_POEPLUS_POWER, True, poe_base.MAX_UPOE_POWER)
        return poe_base.PoeParams(True, 5, ports)

    def adjust_system_poe_power(self, prev_index=-1):
        ''' according to the dni3048up spec, we get

          PSU slots status  |   POE Power Bank Setting      |
          ------------------+-------------------------------+------------------
          PSU2    |   PSU1  |   PB12   |   PB13   |   PB14  | Available Power
          --------+---------+----------+----------+---------+------------------
           1100W  |     x   |      0   |   950W   |     0   |       900W
             x    |  1100W  |      0   |     0    |   900W  |       900W
           1100W  |  1100W  |   1800W  |   900W   |   900W  |      1800W
            715W  |   715W  |   1500W  |   550W   |   550W  |      1200W
           1100W  |   715W  |   1500W  |   900W   |   550W  |      1500W
            715W  |  1100W  |   1500W  |   550W   |   900W  |      1500W
            715W  |     x   |      0   |   550W   |     0   |       550W
              x   |   715W  |      0   |     0    |   550W  |       550W
         ---------+---------+----------+----------+---------+-------------------

        we set up a look-up table to get our values for pb12, pb13 and pb14. If we
        set things up exactly as indicated in the dni3048 spec, we will end up
        disconnecting devices when we go from one power supply to two (because we have
        set PB12 to zero in the case of a single PS and that gets immediately selected
        when a new PS is inserted, resulting in zero power). To account for this,
        we will set PB12 to be equal to the single supply's setting. This way we
        will not lose power. The power will be properly adjusted up when the new supply
        is discovered by poed.

              x         x                                                    0, 0
              x   |   715W  |    550W  |     0    |   550W  |       550W     0, 1
              x   |  1100W  |    900W  |     0    |   900W  |       900W     0, 2
            715W  |     x   |    550W  |   550W   |     0   |       550W     1, 0
            715W  |   715W  |   1200W  |   550W   |   550W  |      1200W     1, 1
            715W  |  1100W  |   1500W  |   550W   |   900W  |      1500W     1, 2
           1100W  |     x   |    900W  |   900W   |     0   |       900W     2, 0
           1100W  |   715W  |   1500W  |   900W   |   550W  |      1500W     2, 1
           1100W  |  1100W  |   1800W  |   900W   |   900W  |      1800W     2, 2
        '''

        self.get_power_bank_settings()
        power_bank_array = [
            (0,    0,   0),    # no power supplies, never used
            (950,  0,   950),
            (950,  0,   950),
            (550,  550, 0),
            (1200, 550, 550),
            (1550, 550, 950),
            (950,  950, 0),
            (1550, 950, 550),
            (1900, 950, 950),
        ]

        power_bank = (12, 13, 14)
        psu1_val = 0
        psu2_val = 0

        # PSU eeproms are not currently readable. For now, we will assume that all
        # PSUs are 1100W
        try:
            if int(open('/sys/bus/i2c/devices/0-0028/psu_pwr1_all_ok').read()) == 1:
                psu1_val = 2
        except:
            pass

        try:
            if int(open('/sys/bus/i2c/devices/0-0028/psu_pwr2_all_ok').read()) == 1:
                psu2_val = 2
        except:
            pass

        '''
        try:
            if int(open('/sys/bus/i2c/devices/0-0028/psu_pwr1_all_ok').read()) == 1:
                inp = open('/sys/bus/i2c/devices/5-0052/eeprom').read()[:5]
                # for now, we are going to assume 1100W for all PSUs
                #psu1_val = 1
                if inp == '1100W':
                    psu1_val = 2
        except:
            pass

        try:
            if int(open('/sys/bus/i2c/devices/0-0028/psu_pwr2_all_ok').read()) == 1:
                inp = open('/sys/bus/i2c/devices/6-0052/eeprom').read()[:5]
                psu2_val = 1
                if inp == '1100W':
                    psu2_val = 2
        except:
            pass
        '''

        index = psu2_val * 3 + psu1_val
        if index == 0:
            return prev_index

        if index != prev_index:
            pba = power_bank_array[index]
            idx = 0
            for power in pba:
                self.set_power_bank(power_bank[idx], power)
                idx += 1
        return index

    def lock(self):
        fcntl.flock(self.fd, fcntl.LOCK_EX)

    def unlock(self):
        fcntl.flock(self.fd, fcntl.LOCK_UN)
